-- scope stuff
qualityoflife = qualityoflife or {};
local q = qualityoflife;

q.modules = q.modules or {};
q.modules["QoL_ClassicThreat-TBC"] = q.modules["QoL_ClassicThreat-TBC"] or {};
local c = q.modules["QoL_ClassicThreat-TBC"];
c.name = "QoL_ClassicThreat-TBC";
c.version = 5; -- check for toc
c.minCoreVersion = 3;

-- main threat cache: memberGuid -> targetGuid -> threatInfo
local threatCache = {};

local threatStatusColors = {
	[0] = { 0.69, 0.69, 0.69 };
	[1] = { 1, 1, 0.47 };
	[2] = { 1, 0.6, 0 };
	[3] = { 1, 0, 0 };
};

-- init
local function UpdateAddon()
	if not QOL_OPTIONS[c.name][q.META_LASTVERSION] or QOL_OPTIONS[c.name][q.META_LASTVERSION] < c.version then
		-- updates here

		QOL_OPTIONS[c.name][q.META_LASTVERSION] = c.version;
	end
	return true;
end

function c:Init()
	if q.initFinished and not c.initFinished and not c.initFailed then
		QOL_OPTIONS = QOL_OPTIONS or {};
		QOL_OPTIONS[c.name] = QOL_OPTIONS[c.name] or {};

		if q.version >= c.minCoreVersion then
			if UpdateAddon() and c:InitOptions() then

				QOL_OPTIONS[c.name]["version"] = c.version;

				c.initFinished = true;
				c:Println(q:GetText("Module loaded."));
				return true;
			end
		else
			c:Println(q:GetText("Error while initialization: QoL Core is outdated (version %s)! Minimum required version: %s", q.version, c.minCoreVersion));
			c:Println(q:GetText("Maybe another of your QoL addon comes with an older version. Please use the latest QoL_Core folder or download the latest version from CurseForge."));
		end
		c.initFailed = true;
	end
end

local function GetThreat(testedUnitId, targetUnitId)
	local isTanking, status, threatpct, rawthreatpct, threatvalue, percentOfLead = UnitDetailedThreatSituation(testedUnitId, targetUnitId);
	if isTanking then
		-- show percentages above 100 if tanking
		percentOfLead = UnitThreatPercentageOfLead(testedUnitId, targetUnitId);
	end
	return isTanking, status, threatpct, rawthreatpct, threatvalue, percentOfLead;
end

local function RemoveMemberFromCache(guid)
	threatCache[guid] = nil;
end

local function RemoveEnemyFromCache(guid)
	for unitGuid, enemyUnitIds in pairs(threatCache) do
		threatCache[unitGuid][guid] = nil;
	end
end

local function GetUnitThreatSituation(testedUnitId)
	local testedUnitGuid = UnitGUID(testedUnitId);
	-- add cache for the tested unitId
	if not threatCache[testedUnitGuid] then
		threatCache[testedUnitGuid] = {};
	end

	-- second cache to not test a threat situation for this unit twice
	local tested = {};

	-- create set of member unitIds (self, allies in party or raid and their pets etc.)
	local memberUnitIds = { "player" };
	if IsInRaid() then
		for playerRaidId, playerName in ipairs(q:GetRaidMembers()[0]) do
			tinsert(memberUnitIds, "raid" .. playerRaidId);
		end
	elseif IsInGroup() then
		for playerPartyId, playerName in pairs(q:GetPartyMembers()) do
			tinsert(memberUnitIds, "party" .. playerPartyId);
		end
	end

	-- loop through member unitIds
	for _, memberUnitId in ipairs(memberUnitIds) do
		if UnitExists(memberUnitId) then
			-- add or remove cache entry for this member unitId
			local memberGuid = UnitGUID(memberUnitId);
			if UnitIsDead(memberUnitId) then
				RemoveMemberFromCache(memberGuid);
			elseif not tested[memberGuid] then
				tested[memberGuid] = {};
			end

			-- create set of potential enemy unitIds based on the member unitId
			local targetUnitIds = {
				memberUnitId .. "target",
				memberUnitId .. "targettarget",
				memberUnitId .. "targettargettarget",
				memberUnitId .. "targettargettargettarget",
				memberUnitId .. "pettarget",
				memberUnitId .. "pettargettarget",
				memberUnitId .. "pettargettargettarget",
				memberUnitId .. "pettargettargettargettarget",
			};

			-- loop through potential enemy unitIds
			for _, targetUnitId in ipairs(targetUnitIds) do
				if UnitExists(targetUnitId) and not UnitIsPlayer(targetUnitId) and not UnitIsFriend(testedUnitId, targetUnitId) then
					local targetGuid = UnitGUID(targetUnitId);

					-- check if this member-enemy combination has already been tested. if not: go on
					if UnitIsDead(targetUnitId) then
						RemoveEnemyFromCache(targetUnitId);
					elseif tested[memberGuid] and not tested[memberGuid][targetGuid] then
						tested[memberGuid][targetGuid] = true;

						-- add threat information to main threat cache
						local isTanking, status, threatpct, rawthreatpct, threatvalue, percentOfLead = GetThreat(testedUnitId, targetUnitId);

						if rawthreatpct or percentOfLead then
							local threatInfo = {};
							threatInfo.targetGuid = targetGuid;
							threatInfo.targetName = UnitName(targetUnitId);
							threatInfo.isTanking = isTanking;
							threatInfo.status = status or 0;
							threatInfo.threatpct = threatpct;
							threatInfo.rawthreatpct = rawthreatpct;
							threatInfo.threatvalue = threatvalue;
							threatInfo.percent = percentOfLead;

							if percentOfLead and percentOfLead > 0 then
								threatInfo.threat = percentOfLead;
							elseif rawthreatpct == 255 then
								threatInfo.threat = threatpct; -- solo workaround
							else
								threatInfo.threat = rawthreatpct;
							end

							if threatCache[testedUnitGuid] then
								-- entry might have been deleted in the mean time due to player death
								threatCache[testedUnitGuid][targetGuid] = threatInfo;
							end
						end
					end
				end
			end
		end
	end

	-- get highest value from all checked combinations for the tested unitId and return
	local highestStatusThreatInfo, highestThreatPctThreatInfo;
	if threatCache[testedUnitGuid] then
		for targetGuid, threatInfo in pairs(threatCache[testedUnitGuid]) do
			if not (testedUnitId == "player" and UnitGUID("target") == targetGuid) then -- dont show target threat on PlayerFrame if already shown on TargetFrame
				if not highestStatusThreatInfo or threatInfo.status > highestStatusThreatInfo.status then
					highestStatusThreatInfo = threatInfo;
				end
				if not highestThreatPctThreatInfo or threatInfo.threat > highestThreatPctThreatInfo.threat then
					highestThreatPctThreatInfo = threatInfo;
				end
			end
		end
	end

	return highestStatusThreatInfo, highestThreatPctThreatInfo;
end

local function ShowTargetThreat()
	if TargetFrame:IsVisible() then

		local frame = c:GetTargetThreatFrame();
		if UnitExists("target") and not UnitIsPlayer("target") and not UnitIsFriend("player", "target") then

			local isTanking, status, threatpct, rawthreatpct, threatvalue, percentOfLead = GetThreat("player", "target");
			status = status or 0;
			rawthreatpct = rawthreatpct or 0;
			if percentOfLead then
				rawthreatpct = percentOfLead;
			end

			if status and rawthreatpct > 0 then
				frame.text:SetText(q:Round(rawthreatpct, 0) .. "%");
				if status > 0 then
					frame.background:SetVertexColor(unpack(threatStatusColors[status]));
					frame.background:Show();
					frame.glow:SetVertexColor(unpack(threatStatusColors[status]));
					frame.glow:Show();
				else
					frame.background:Hide();
					frame.glow:Hide();
				end
				frame:Show();
				return;
			end
		end
		if frame:IsVisible() then
			frame:Hide();
		end
	end
end

local function ShowPlayerThreat()
	local frame = c:GetPlayerThreatFrame();
	local highestStatusThreatInfo, highestThreatPctThreatInfo = GetUnitThreatSituation("player");
	local status, threat = 0, 0;
	if highestStatusThreatInfo then
		status = highestStatusThreatInfo.status or status;
	end
	if highestThreatPctThreatInfo then
		threat = highestThreatPctThreatInfo.threat or threat;
	end

	if status and threat > 0 then
		frame.text:SetText(q:Round(threat, 0) .. "%");
		if status > 0 then
			frame.background:SetVertexColor(unpack(threatStatusColors[status]));
			frame.background:Show();
		else
			frame.background:Hide();
		end
		frame:Show();
		return;
	end
	frame:Hide();
end

local function ShowPartyThreat(playerPartyId)
	local unitId = "party" .. playerPartyId;
	local frame = c:GetPartyFrame(playerPartyId);
	if UnitExists(unitId) and not UnitIsDead(unitId) then
		local highestStatusThreatInfo, highestThreatPctThreatInfo = GetUnitThreatSituation(unitId);
		local status, threat = 0, 0;
		if highestStatusThreatInfo and highestStatusThreatInfo.status then
			status = highestStatusThreatInfo.status;
		end
		if highestThreatPctThreatInfo and highestThreatPctThreatInfo.threat then
			threat = highestStatusThreatInfo.threat;
		end
		if status > 0 or threat > 0 then
			frame.text:SetText(q:Round(threat, 0) .. "%");
			if status > 0 then
				frame.background:SetVertexColor(unpack(threatStatusColors[status]));
				frame.background:Show();
			else
				frame.background:Hide();
			end
			frame:Show();
			return;
		end
	end
	frame:Hide();
end

-- TODO
local function ShowRaidThreat(raidPlayerId)
	local unitId = "raid" .. raidPlayerId;
	local frame = c:GetRaidFrame(raidPlayerId);
	if UnitExists(unitId) and not UnitIsDead(unitId) then
		local highestStatus, highestThreatPct = GetUnitThreatSituation(unitId);
		if highestStatus and highestThreatPct then
			frame.text:SetText(q:Round(highestThreatPct, 0) .. "%");
			if highestStatus > 0 then
				frame.background:SetVertexColor(unpack(threatStatusColors[highestStatus]));
				frame.background:Show();
			else
				frame.background:Hide();
			end
			frame:Show();
		elseif frame:IsVisible() then
			frame:Hide();
		end
	end
end

local function UpdateThreat()
	ShowPlayerThreat();
	ShowTargetThreat();
	for playerPartyId, playerName in pairs(q:GetPartyMembers()) do
		ShowPartyThreat(playerPartyId);
	end
	--TODO: RAID
end

local function Reset()
	threatCache = {};
	c:HideFrames();
end

local function OnCombatEvent(event, ...)
	local timestamp, subevent, _, sourceGUID, sourceName, sourceFlags, sourceRaidFlags, destGUID, destName, destFlags, destRaidFlags = ...;
	if subevent == "UNIT_DIED" and threatCache then
		-- timer functions are neccessary because async threat server methods have noticable delay
		if destGUID == UnitGUID("player") then
			-- player dies
			Reset();
			C_Timer.After(1, function ()
				Reset();
			end);
		elseif q:StartsWith(destGUID, "Player") then
			-- might have been a member who died
			if IsInGroup() or IsInRaid() then
				RemoveMemberFromCache(destGUID);
				C_Timer.After(1, function ()
					RemoveMemberFromCache(destGUID);
				end);
			end
		else
			-- enemy died: remove from cache
			RemoveEnemyFromCache(destGUID);
			UpdateThreat();
			C_Timer.After(1, function ()
				RemoveEnemyFromCache(destGUID);
				UpdateThreat();
			end);
		end
	end
end

local eventFrame;
local function HookEvents()
	if not eventFrame then
		eventFrame = q:GetEventFrame();
		eventFrame:RegisterEvent("UNIT_THREAT_LIST_UPDATE");
		eventFrame:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED");
		eventFrame:RegisterEvent("PLAYER_REGEN_ENABLED");
		eventFrame:RegisterEvent("PLAYER_REGEN_DISABLED");
		eventFrame:RegisterEvent("PLAYER_TARGET_CHANGED");
		eventFrame:RegisterEvent("PLAYER_ENTERING_WORLD");
		eventFrame:HookScript("OnEvent", function(self, event, arg1, arg2, arg3, arg4)
			if event then
				if event == "ADDON_LOADED" and arg1 == c.name then
					c:Init();
				elseif c.initFinished then
					if event == "PLAYER_ENTERING_WORLD" then
						Reset();
					elseif event == "PLAYER_REGEN_ENABLED" then
						Reset();
					elseif event == "PLAYER_REGEN_DISABLED" then
						Reset();
					elseif event == "COMBAT_LOG_EVENT_UNFILTERED" then
						OnCombatEvent(event, CombatLogGetCurrentEventInfo());
					elseif event == "UNIT_THREAT_LIST_UPDATE" then
						UpdateThreat();
					elseif event == "PLAYER_TARGET_CHANGED" then
						ShowTargetThreat();
						ShowPlayerThreat();
					end
				end
			end
		end);
	end
end
HookEvents();